//! This file has been automatically generated by `objc2`'s `header-translator`.
//! DO NOT EDIT
use objc2::__framework_prelude::*;

use crate::*;

/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilenosuchfileerror?language=objc)
pub const NSFileNoSuchFileError: NSInteger = 4;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilelockingerror?language=objc)
pub const NSFileLockingError: NSInteger = 255;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilereadunknownerror?language=objc)
pub const NSFileReadUnknownError: NSInteger = 256;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilereadnopermissionerror?language=objc)
pub const NSFileReadNoPermissionError: NSInteger = 257;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilereadinvalidfilenameerror?language=objc)
pub const NSFileReadInvalidFileNameError: NSInteger = 258;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilereadcorruptfileerror?language=objc)
pub const NSFileReadCorruptFileError: NSInteger = 259;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilereadnosuchfileerror?language=objc)
pub const NSFileReadNoSuchFileError: NSInteger = 260;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilereadinapplicablestringencodingerror?language=objc)
pub const NSFileReadInapplicableStringEncodingError: NSInteger = 261;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilereadunsupportedschemeerror?language=objc)
pub const NSFileReadUnsupportedSchemeError: NSInteger = 262;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilereadtoolargeerror?language=objc)
pub const NSFileReadTooLargeError: NSInteger = 263;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilereadunknownstringencodingerror?language=objc)
pub const NSFileReadUnknownStringEncodingError: NSInteger = 264;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilewriteunknownerror?language=objc)
pub const NSFileWriteUnknownError: NSInteger = 512;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilewritenopermissionerror?language=objc)
pub const NSFileWriteNoPermissionError: NSInteger = 513;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilewriteinvalidfilenameerror?language=objc)
pub const NSFileWriteInvalidFileNameError: NSInteger = 514;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilewritefileexistserror?language=objc)
pub const NSFileWriteFileExistsError: NSInteger = 516;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilewriteinapplicablestringencodingerror?language=objc)
pub const NSFileWriteInapplicableStringEncodingError: NSInteger = 517;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilewriteunsupportedschemeerror?language=objc)
pub const NSFileWriteUnsupportedSchemeError: NSInteger = 518;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilewriteoutofspaceerror?language=objc)
pub const NSFileWriteOutOfSpaceError: NSInteger = 640;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilewritevolumereadonlyerror?language=objc)
pub const NSFileWriteVolumeReadOnlyError: NSInteger = 642;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilemanagerunmountunknownerror?language=objc)
pub const NSFileManagerUnmountUnknownError: NSInteger = 768;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfilemanagerunmountbusyerror?language=objc)
pub const NSFileManagerUnmountBusyError: NSInteger = 769;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nskeyvaluevalidationerror?language=objc)
pub const NSKeyValueValidationError: NSInteger = 1024;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsformattingerror?language=objc)
pub const NSFormattingError: NSInteger = 2048;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsusercancellederror?language=objc)
pub const NSUserCancelledError: NSInteger = 3072;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfeatureunsupportederror?language=objc)
pub const NSFeatureUnsupportedError: NSInteger = 3328;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsexecutablenotloadableerror?language=objc)
pub const NSExecutableNotLoadableError: NSInteger = 3584;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsexecutablearchitecturemismatcherror?language=objc)
pub const NSExecutableArchitectureMismatchError: NSInteger = 3585;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsexecutableruntimemismatcherror?language=objc)
pub const NSExecutableRuntimeMismatchError: NSInteger = 3586;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsexecutableloaderror?language=objc)
pub const NSExecutableLoadError: NSInteger = 3587;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsexecutablelinkerror?language=objc)
pub const NSExecutableLinkError: NSInteger = 3588;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfileerrorminimum?language=objc)
pub const NSFileErrorMinimum: NSInteger = 0;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsfileerrormaximum?language=objc)
pub const NSFileErrorMaximum: NSInteger = 1023;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsvalidationerrorminimum?language=objc)
pub const NSValidationErrorMinimum: NSInteger = 1024;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsvalidationerrormaximum?language=objc)
pub const NSValidationErrorMaximum: NSInteger = 2047;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsexecutableerrorminimum?language=objc)
pub const NSExecutableErrorMinimum: NSInteger = 3584;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsexecutableerrormaximum?language=objc)
pub const NSExecutableErrorMaximum: NSInteger = 3839;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsformattingerrorminimum?language=objc)
pub const NSFormattingErrorMinimum: NSInteger = 2048;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsformattingerrormaximum?language=objc)
pub const NSFormattingErrorMaximum: NSInteger = 2559;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nspropertylistreadcorrupterror?language=objc)
pub const NSPropertyListReadCorruptError: NSInteger = 3840;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nspropertylistreadunknownversionerror?language=objc)
pub const NSPropertyListReadUnknownVersionError: NSInteger = 3841;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nspropertylistreadstreamerror?language=objc)
pub const NSPropertyListReadStreamError: NSInteger = 3842;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nspropertylistwritestreamerror?language=objc)
pub const NSPropertyListWriteStreamError: NSInteger = 3851;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nspropertylistwriteinvaliderror?language=objc)
pub const NSPropertyListWriteInvalidError: NSInteger = 3852;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nspropertylisterrorminimum?language=objc)
pub const NSPropertyListErrorMinimum: NSInteger = 3840;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nspropertylisterrormaximum?language=objc)
pub const NSPropertyListErrorMaximum: NSInteger = 4095;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsxpcconnectioninterrupted?language=objc)
pub const NSXPCConnectionInterrupted: NSInteger = 4097;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsxpcconnectioninvalid?language=objc)
pub const NSXPCConnectionInvalid: NSInteger = 4099;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsxpcconnectionreplyinvalid?language=objc)
pub const NSXPCConnectionReplyInvalid: NSInteger = 4101;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsxpcconnectioncodesigningrequirementfailure?language=objc)
pub const NSXPCConnectionCodeSigningRequirementFailure: NSInteger = 4102;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsxpcconnectionerrorminimum?language=objc)
pub const NSXPCConnectionErrorMinimum: NSInteger = 4096;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsxpcconnectionerrormaximum?language=objc)
pub const NSXPCConnectionErrorMaximum: NSInteger = 4224;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsubiquitousfileunavailableerror?language=objc)
pub const NSUbiquitousFileUnavailableError: NSInteger = 4353;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsubiquitousfilenotuploadedduetoquotaerror?language=objc)
pub const NSUbiquitousFileNotUploadedDueToQuotaError: NSInteger = 4354;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsubiquitousfileubiquityservernotavailable?language=objc)
pub const NSUbiquitousFileUbiquityServerNotAvailable: NSInteger = 4355;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsubiquitousfileerrorminimum?language=objc)
pub const NSUbiquitousFileErrorMinimum: NSInteger = 4352;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsubiquitousfileerrormaximum?language=objc)
pub const NSUbiquitousFileErrorMaximum: NSInteger = 4607;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsuseractivityhandofffailederror?language=objc)
pub const NSUserActivityHandoffFailedError: NSInteger = 4608;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsuseractivityconnectionunavailableerror?language=objc)
pub const NSUserActivityConnectionUnavailableError: NSInteger = 4609;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsuseractivityremoteapplicationtimedouterror?language=objc)
pub const NSUserActivityRemoteApplicationTimedOutError: NSInteger = 4610;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsuseractivityhandoffuserinfotoolargeerror?language=objc)
pub const NSUserActivityHandoffUserInfoTooLargeError: NSInteger = 4611;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsuseractivityerrorminimum?language=objc)
pub const NSUserActivityErrorMinimum: NSInteger = 4608;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsuseractivityerrormaximum?language=objc)
pub const NSUserActivityErrorMaximum: NSInteger = 4863;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscoderreadcorrupterror?language=objc)
pub const NSCoderReadCorruptError: NSInteger = 4864;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscodervaluenotfounderror?language=objc)
pub const NSCoderValueNotFoundError: NSInteger = 4865;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscoderinvalidvalueerror?language=objc)
pub const NSCoderInvalidValueError: NSInteger = 4866;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscodererrorminimum?language=objc)
pub const NSCoderErrorMinimum: NSInteger = 4864;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscodererrormaximum?language=objc)
pub const NSCoderErrorMaximum: NSInteger = 4991;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsbundleerrorminimum?language=objc)
pub const NSBundleErrorMinimum: NSInteger = 4992;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsbundleerrormaximum?language=objc)
pub const NSBundleErrorMaximum: NSInteger = 5119;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsbundleondemandresourceoutofspaceerror?language=objc)
pub const NSBundleOnDemandResourceOutOfSpaceError: NSInteger = 4992;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsbundleondemandresourceexceededmaximumsizeerror?language=objc)
pub const NSBundleOnDemandResourceExceededMaximumSizeError: NSInteger = 4993;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsbundleondemandresourceinvalidtagerror?language=objc)
pub const NSBundleOnDemandResourceInvalidTagError: NSInteger = 4994;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscloudsharingnetworkfailureerror?language=objc)
pub const NSCloudSharingNetworkFailureError: NSInteger = 5120;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscloudsharingquotaexceedederror?language=objc)
pub const NSCloudSharingQuotaExceededError: NSInteger = 5121;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscloudsharingtoomanyparticipantserror?language=objc)
pub const NSCloudSharingTooManyParticipantsError: NSInteger = 5122;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscloudsharingconflicterror?language=objc)
pub const NSCloudSharingConflictError: NSInteger = 5123;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscloudsharingnopermissionerror?language=objc)
pub const NSCloudSharingNoPermissionError: NSInteger = 5124;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscloudsharingothererror?language=objc)
pub const NSCloudSharingOtherError: NSInteger = 5375;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscloudsharingerrorminimum?language=objc)
pub const NSCloudSharingErrorMinimum: NSInteger = 5120;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscloudsharingerrormaximum?language=objc)
pub const NSCloudSharingErrorMaximum: NSInteger = 5375;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscompressionfailederror?language=objc)
pub const NSCompressionFailedError: NSInteger = 5376;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsdecompressionfailederror?language=objc)
pub const NSDecompressionFailedError: NSInteger = 5377;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscompressionerrorminimum?language=objc)
pub const NSCompressionErrorMinimum: NSInteger = 5376;
/// [Apple's documentation](https://developer.apple.com/documentation/foundation/nscompressionerrormaximum?language=objc)
pub const NSCompressionErrorMaximum: NSInteger = 5503;
