//! This file has been automatically generated by `objc2`'s `header-translator`.
//! DO NOT EDIT
use core::ffi::*;

use crate::*;

/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nstextreadinapplicabledocumenttypeerror?language=objc)
pub const NSTextReadInapplicableDocumentTypeError: c_uint = 65806;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nstextwriteinapplicabledocumenttypeerror?language=objc)
pub const NSTextWriteInapplicableDocumentTypeError: c_uint = 66062;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nstextreadwriteerrorminimum?language=objc)
pub const NSTextReadWriteErrorMinimum: c_uint = 65792;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nstextreadwriteerrormaximum?language=objc)
pub const NSTextReadWriteErrorMaximum: c_uint = 66303;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsfontassetdownloaderror?language=objc)
pub const NSFontAssetDownloadError: c_uint = 66304;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsfonterrorminimum?language=objc)
pub const NSFontErrorMinimum: c_uint = 66304;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsfonterrormaximum?language=objc)
pub const NSFontErrorMaximum: c_uint = 66335;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsserviceapplicationnotfounderror?language=objc)
pub const NSServiceApplicationNotFoundError: c_uint = 66560;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsserviceapplicationlaunchfailederror?language=objc)
pub const NSServiceApplicationLaunchFailedError: c_uint = 66561;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsservicerequesttimedouterror?language=objc)
pub const NSServiceRequestTimedOutError: c_uint = 66562;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsserviceinvalidpasteboarddataerror?language=objc)
pub const NSServiceInvalidPasteboardDataError: c_uint = 66563;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsservicemalformedservicedictionaryerror?language=objc)
pub const NSServiceMalformedServiceDictionaryError: c_uint = 66564;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsservicemiscellaneouserror?language=objc)
pub const NSServiceMiscellaneousError: c_uint = 66800;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsserviceerrorminimum?language=objc)
pub const NSServiceErrorMinimum: c_uint = 66560;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsserviceerrormaximum?language=objc)
pub const NSServiceErrorMaximum: c_uint = 66817;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nssharingservicenotconfigurederror?language=objc)
pub const NSSharingServiceNotConfiguredError: c_uint = 67072;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nssharingserviceerrorminimum?language=objc)
pub const NSSharingServiceErrorMinimum: c_uint = 67072;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nssharingserviceerrormaximum?language=objc)
pub const NSSharingServiceErrorMaximum: c_uint = 67327;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsworkspaceauthorizationinvaliderror?language=objc)
pub const NSWorkspaceAuthorizationInvalidError: c_uint = 67328;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsworkspaceerrorminimum?language=objc)
pub const NSWorkspaceErrorMinimum: c_uint = 67328;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nsworkspaceerrormaximum?language=objc)
pub const NSWorkspaceErrorMaximum: c_uint = 67455;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nswindowsharingrequestalreadyrequested?language=objc)
pub const NSWindowSharingRequestAlreadyRequested: c_uint = 67456;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nswindowsharingrequestnoeligiblesession?language=objc)
pub const NSWindowSharingRequestNoEligibleSession: c_uint = 67457;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nswindowsharingrequestunspecifiederror?language=objc)
pub const NSWindowSharingRequestUnspecifiedError: c_uint = 67458;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nswindowsharingerrorminimum?language=objc)
pub const NSWindowSharingErrorMinimum: c_uint = 67456;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nswindowsharingerrormaximum?language=objc)
pub const NSWindowSharingErrorMaximum: c_uint = 67466;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nspasteboardmiscellaneouserror?language=objc)
pub const NSPasteboardMiscellaneousError: c_uint = 67584;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nspasteboardcommunicationerror?language=objc)
pub const NSPasteboardCommunicationError: c_uint = 67585;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nspasteboardinvalidargumenterror?language=objc)
pub const NSPasteboardInvalidArgumentError: c_uint = 67586;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nspasteboardcontentsnotavailableerror?language=objc)
pub const NSPasteboardContentsNotAvailableError: c_uint = 67587;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nspasteboarderrorminimum?language=objc)
pub const NSPasteboardErrorMinimum: c_uint = 67584;
/// [Apple's documentation](https://developer.apple.com/documentation/appkit/nspasteboarderrormaximum?language=objc)
pub const NSPasteboardErrorMaximum: c_uint = 67839;
