//! This file has been automatically generated by `objc2`'s `header-translator`.
//! DO NOT EDIT
use core::ptr::NonNull;
use objc2::__framework_prelude::*;
#[cfg(feature = "objc2-core-foundation")]
use objc2_core_foundation::*;

use crate::*;

extern_class!(
    /// [Apple's documentation](https://developer.apple.com/documentation/foundation/nsnull?language=objc)
    #[unsafe(super(NSObject))]
    #[derive(Debug, PartialEq, Eq, Hash)]
    pub struct NSNull;
);

unsafe impl Send for NSNull {}

unsafe impl Sync for NSNull {}

#[cfg(feature = "objc2-core-foundation")]
impl AsRef<NSNull> for CFNull {
    #[inline]
    fn as_ref(&self) -> &NSNull {
        unsafe { &*((self as *const Self).cast()) }
    }
}

#[cfg(feature = "objc2-core-foundation")]
impl AsRef<CFNull> for NSNull {
    #[inline]
    fn as_ref(&self) -> &CFNull {
        unsafe { &*((self as *const Self).cast()) }
    }
}

#[cfg(feature = "NSObject")]
extern_conformance!(
    unsafe impl NSCoding for NSNull {}
);

#[cfg(feature = "NSObject")]
extern_conformance!(
    unsafe impl NSCopying for NSNull {}
);

#[cfg(feature = "NSObject")]
unsafe impl CopyingHelper for NSNull {
    type Result = Self;
}

extern_conformance!(
    unsafe impl NSObjectProtocol for NSNull {}
);

#[cfg(feature = "NSObject")]
extern_conformance!(
    unsafe impl NSSecureCoding for NSNull {}
);

impl NSNull {
    extern_methods!(
        #[unsafe(method(null))]
        #[unsafe(method_family = none)]
        pub fn null() -> Retained<NSNull>;
    );
}

/// Methods declared on superclass `NSObject`.
impl NSNull {
    extern_methods!(
        #[unsafe(method(init))]
        #[unsafe(method_family = init)]
        pub fn init(this: Allocated<Self>) -> Retained<Self>;

        #[unsafe(method(new))]
        #[unsafe(method_family = new)]
        pub fn new() -> Retained<Self>;
    );
}

impl DefaultRetained for NSNull {
    #[inline]
    fn default_retained() -> Retained<Self> {
        Self::new()
    }
}
